import itertools
from pathlib import Path
from shutil import rmtree

import click
import questionary as q

from vantage6.common import info
from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import InstanceType
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.context import ServerContext
from vantage6.cli.utils import remove_file


@click.command()
@click_insert_context(type_=InstanceType.SERVER)
@click.option("-f", "--force", "force", flag_value=True)
def cli_server_remove(ctx: ServerContext, force: bool) -> None:
    """
    Function to remove a server.

    Parameters
    ----------
    ctx : ServerContext
        Server context object
    force : bool
        Whether to ask for confirmation before removing or not
    """
    check_docker_running()

    if not force:
        if not q.confirm(
            "This server will be deleted permanently including its "
            "configuration. Are you sure?",
            default=False,
        ).ask():
            info("Server will not be deleted")
            exit(0)

    # now remove the folders...
    remove_file(ctx.config_file, "configuration")

    # ensure log files are closed before removing
    log_dir = Path(ctx.log_file.parent)
    if log_dir.exists():
        info(f"Removing log directory: {log_dir}")
        for handler in itertools.chain(ctx.log.handlers, ctx.log.root.handlers):
            handler.close()
        # remove the whole folder with all the log files. This may also still contain other
        # files like RabbitMQ configuration etc
        rmtree(log_dir)



import click
import docker
import questionary as q

from colorama import Fore, Style

from vantage6.common import error
from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import APPNAME, InstanceType
from vantage6.cli.common.start import attach_logs
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS


@click.command()
@click.option("-n", "--name", default=None, help="configuration name")
@click.option("--system", "system_folders", flag_value=True)
@click.option(
    "--user", "system_folders", flag_value=False, default=DEFAULT_SERVER_SYSTEM_FOLDERS
)
def cli_algo_store_attach(name: str, system_folders: bool) -> None:
    """
    Show the server logs in the current console.
    """
    check_docker_running()
    client = docker.from_env()

    running_servers = client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.ALGORITHM_STORE.value}"}
    )
    running_server_names = [container.name for container in running_servers]

    if not name:
        try:
            name = q.select(
                "Select the algorithm store you wish to attach:",
                choices=running_server_names,
            ).unsafe_ask()
        except KeyboardInterrupt:
            error("Aborted by user!")
            return
    else:
        post_fix = "system" if system_folders else "user"
        name = f"{APPNAME}-{name}-{post_fix}-{InstanceType.ALGORITHM_STORE.value}"

    if name in running_server_names:
        container = client.containers.get(name)
        attach_logs(container, InstanceType.ALGORITHM_STORE)
    else:
        error(f"{Fore.RED}{name}{Style.RESET_ALL} is not running!")



import os
from threading import Thread

import click
import docker
from sqlalchemy.engine.url import make_url
from vantage6.cli.globals import ServerGlobals

from vantage6.common import info, warning
from vantage6.common.docker.addons import check_docker_running, pull_image
from vantage6.common.globals import (
    APPNAME,
    DEFAULT_DOCKER_REGISTRY,
    DEFAULT_SERVER_IMAGE,
    InstanceType,
)
from vantage6.cli.context.server import ServerContext
from vantage6.cli.utils import check_config_name_allowed
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.utils import print_log_worker


# TODO this method has a lot of duplicated code from `start`
@click.command()
@click.argument("file", type=click.Path(exists=True))
@click.option(
    "--drop-all",
    is_flag=True,
    default=False,
    help="Drop all existing data before importing",
)
@click.option("-i", "--image", default=None, help="Node Docker image to use")
@click.option(
    "--mount-src",
    default="",
    help="Override vantage6 source code in container with the source"
    " code in this path",
)
@click.option(
    "--keep/--auto-remove",
    default=False,
    help="Keep image after finishing. Useful for debugging",
)
@click.option("--wait", default=False, help="Wait for the import to finish")
@click_insert_context(type_="server")
def cli_server_import(
    ctx: ServerContext,
    file: str,
    drop_all: bool,
    image: str | None,
    mount_src: str,
    keep: bool,
    wait: bool,
) -> None:
    """
    Import vantage6 resources into a server instance.

    This allows you to create organizations, collaborations, users, tasks, etc
    from a yaml file.

    The FILE_ argument should be a path to a yaml file containing the vantage6
    formatted data to import.
    """
    # will print an error if not
    check_docker_running()

    info("Starting server...")
    info("Finding Docker daemon.")
    docker_client = docker.from_env()

    # check if name is allowed for docker volume, else exit
    check_config_name_allowed(ctx.name)

    # pull latest Docker image
    if image is None:
        image = ctx.config.get(
            "image", f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_SERVER_IMAGE}"
        )
    info(f"Pulling latest server image '{image}'.")
    pull_image(docker_client, image)

    info("Creating mounts")
    mounts = [
        docker.types.Mount("/mnt/config.yaml", str(ctx.config_file), type="bind"),
        docker.types.Mount("/mnt/import.yaml", str(file), type="bind"),
    ]

    # FIXME: code duplication with cli_server_start()
    # try to mount database
    uri = ctx.config["uri"]
    url = make_url(uri)
    environment_vars = None

    if mount_src:
        mount_src = os.path.abspath(mount_src)
        mounts.append(docker.types.Mount("/vantage6", mount_src, type="bind"))

    # If host is None, we're dealing with a file-based DB, like SQLite
    if url.host is None:
        db_path = url.database

        if not os.path.isabs(db_path):
            # We're dealing with a relative path here -> make it absolute
            db_path = ctx.data_dir / url.database

        basename = os.path.basename(db_path)
        dirname = os.path.dirname(db_path)
        os.makedirs(dirname, exist_ok=True)

        # we're mounting the entire folder that contains the database
        mounts.append(docker.types.Mount("/mnt/database/", dirname, type="bind"))

        environment_vars = {
            ServerGlobals.DB_URI_ENV_VAR: f"sqlite:////mnt/database/{basename}"
        }

    else:
        warning(
            f"Database could not be transferred, make sure {url.host} "
            "is reachable from the Docker container"
        )
        info("Consider using the docker-compose method to start a server")

    drop_all_ = "--drop-all" if drop_all else ""
    cmd = f"vserver-local import -c /mnt/config.yaml {drop_all_} /mnt/import.yaml"

    info(cmd)

    info("Run Docker container")
    container = docker_client.containers.run(
        image,
        command=cmd,
        mounts=mounts,
        detach=True,
        labels={
            f"{APPNAME}-type": InstanceType.SERVER.value,
            "name": ctx.config_file_name,
        },
        environment=environment_vars,
        auto_remove=not keep,
        tty=True,
    )
    logs = container.logs(stream=True, stdout=True)
    Thread(target=print_log_worker, args=(logs,), daemon=False).start()

    info(f"Success! container id = {container.id}")

    if wait:
        container.wait()
        info("Container finished!")



from __future__ import annotations

import os.path

from sqlalchemy.engine.url import make_url

from vantage6.common.context import AppContext
from vantage6.cli.globals import (
    DEFAULT_SERVER_SYSTEM_FOLDERS as S_FOL,
    ServerType,
)


class BaseServerContext(AppContext):
    """
    Base context for a vantage6 server or algorithm store server

    Contains functions that the ServerContext and AlgorithmStoreContext have
    in common.
    """

    def get_database_uri(self, db_env_var: str) -> str:
        """
        Obtain the database uri from the environment or the configuration.

        Parameters
        ----------
        db_env_var : str
            Name of the environment variable that contains the database uri

        Returns
        -------
        str
            string representation of the database uri
        """
        uri = os.environ.get(db_env_var) or self.config["uri"]
        url = make_url(uri)

        if url.host is None and not os.path.isabs(url.database):
            # We're dealing with a relative path here of a local database, when
            # we're running the server outside of docker. Therefore we need to
            # prepend the data directory to the database name, but after the
            # driver name (e.g. sqlite:////db.sqlite ->
            # sqlite:////data_dir>/db.sqlite)

            # find index of database name
            idx_db_name = str(url).find(url.database)

            # add the datadir to the right location in the database uri
            return str(url)[:idx_db_name] + str(self.data_dir / url.database)

        return uri

    @classmethod
    def from_external_config_file(
        cls,
        path: str,
        server_type: ServerType,
        config_name_env_var: str,
        system_folders: bool = S_FOL,
    ) -> BaseServerContext:
        """
        Create a server context from an external configuration file. External
        means that the configuration file is not located in the default folders
        but its location is specified by the user.

        Parameters
        ----------
        path : str
            Path of the configuration file
        server_type : ServerType
            Type of server, either 'server' or 'algorithm-store'
        config_name_env_var : str
            Name of the environment variable that contains the name of the
            configuration
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL

        Returns
        -------
        ServerContext
            Server context object
        """
        cls = super().from_external_config_file(path, server_type, system_folders)
        # if we are running a server in a docker container, the name is taken
        # from the name of the config file (which is usually a default). Get
        # the config name from environment if it is given.
        cls.name = os.environ.get(config_name_env_var) or cls.name
        return cls



import subprocess
import itertools
from shutil import rmtree
from pathlib import Path

import click
import docker
from colorama import Fore, Style

from vantage6.cli.context.algorithm_store import AlgorithmStoreContext
from vantage6.common import info, error
from vantage6.common.globals import APPNAME
from vantage6.cli.context.server import ServerContext
from vantage6.cli.context.node import NodeContext
from vantage6.cli.server.remove import cli_server_remove
from vantage6.cli.utils import remove_file
from vantage6.common.globals import InstanceType
from vantage6.cli.dev.utils import get_dev_server_context


@click.command()
@click.option("-n", "--name", default=None, help="Name of the configuration.")
@click.option(
    "-c",
    "--config",
    default=None,
    help="Path to configuration-file; overrides --name",
)
@click.pass_context
def remove_demo_network(
    click_ctx: click.Context, name: str | None, config: str | None
) -> None:
    """Remove all related demo network files and folders.

    Select a server configuration to remove that server and the nodes attached
    to it.
    """
    ctx = get_dev_server_context(config, name)

    # check that the server is not running
    client = docker.from_env()
    running_servers = client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.SERVER.value}"}
    )
    running_server_names = [server.name for server in running_servers]
    container_name = f"{APPNAME}-{name}-user-{InstanceType.SERVER.value}"
    if container_name in running_server_names:
        error(
            f"Server {Fore.RED}{name}{Style.RESET_ALL} is still running! First stop "
            "the network with 'v6 dev stop-demo-network'."
        )
        return

    # remove the server
    for handler in itertools.chain(ctx.log.handlers, ctx.log.root.handlers):
        handler.close()
    click_ctx.invoke(cli_server_remove, ctx=ctx, force=True)

    # removing the server import config
    info("Deleting demo import config file")
    server_configs = ServerContext.instance_folders(
        InstanceType.SERVER, ctx.name, system_folders=False
    )
    import_config_to_del = Path(server_configs["dev"]) / f"{ctx.name}.yaml"
    remove_file(import_config_to_del, "import_configuration")

    # also remove the server folder
    server_configs = ServerContext.instance_folders(
        InstanceType.SERVER, ctx.name, system_folders=False
    )
    server_folder = server_configs["data"]
    if server_folder.is_dir():
        rmtree(server_folder)

    # remove the store folder
    store_configs = AlgorithmStoreContext.instance_folders(
        InstanceType.ALGORITHM_STORE, f"{ctx.name}_store", system_folders=False
    )
    store_folder = store_configs["data"]
    if store_folder.is_dir():
        rmtree(store_folder)

    # remove the store config file
    subprocess.run(
        [
            "v6",
            "algorithm-store",
            "remove",
            "-n",
            f"{ctx.name}_store",
            "--force",
            "--user",
        ]
    )

    # remove the nodes
    configs, _ = NodeContext.available_configurations(system_folders=False)
    node_names = [
        config.name for config in configs if config.name.startswith(f"{ctx.name}_node_")
    ]
    for name in node_names:
        node_ctx = NodeContext(name, False)
        for handler in itertools.chain(
            node_ctx.log.handlers, node_ctx.log.root.handlers
        ):
            handler.close()
        subprocess.run(["v6", "node", "remove", "-n", name, "--user", "--force"])

    # remove data files attached to the network
    data_dirs_nodes = NodeContext.instance_folders("node", "", False)["dev"]
    rmtree(Path(data_dirs_nodes / ctx.name))



"""
This module contains global variables that are used throughout the CLI.
"""

from enum import Enum
from pathlib import Path
from vantage6.common.globals import APPNAME

#
#   SERVER SETTINGS
#
DEFAULT_SERVER_SYSTEM_FOLDERS = True

#
#   NODE SETTINGS
#
DEFAULT_NODE_SYSTEM_FOLDERS = False

#
#   INSTALLATION SETTINGS
#
PACKAGE_FOLDER = Path(__file__).parent.parent.parent

# FIXME BvB 22-06-28 think this is also defined in the node globals, and this
# one appears not to be used
NODE_PROXY_SERVER_HOSTNAME = "proxyserver"

DATA_FOLDER = PACKAGE_FOLDER / APPNAME / "_data"

# with open(Path(PACKAGE_FOLDER) / APPNAME / "cli" / "VERSION") as f:
#     VERSION = f.read()

# Maximum time to start up RabbitMQ in seconds
RABBIT_TIMEOUT = 300

# Location of repository to create new algorithm templates from
ALGORITHM_TEMPLATE_REPO = "gh:vantage6/v6-algorithm-template.git"

# image to use for diagnostics in `v6 test` commands
DIAGNOSTICS_IMAGE = "harbor2.vantage6.ai/algorithms/diagnostic"

# Address of community algorithm store
COMMUNITY_STORE = "https://store.cotopaxi.vantage6.ai/api"

DEFAULT_PROMETHEUS_IMAGE = "prom/prometheus"
PROMETHEUS_CONFIG = "prometheus.yml"
PROMETHEUS_DIR = "prometheus"


# datasets included in the nodes of the dev network
class DefaultDatasets(str, Enum):
    """Enum containing default datasets"""

    OLYMPIC_ATHLETES = "olympic_athletes_2016.csv"
    KAPLAN_MEIER_TEST = "km_dataset.csv"


class ServerType(str, Enum):
    """Enum containing server types"""

    V6SERVER = "server"
    ALGORITHM_STORE = "algorithm-store"


class ServerGlobals(str, Enum):
    """Enum containing server environment variables"""

    DB_URI_ENV_VAR = "VANTAGE6_DB_URI"
    CONFIG_NAME_ENV_VAR = "VANTAGE6_CONFIG_NAME"


class AlgoStoreGlobals(str, Enum):
    """Enum containing algorithm store environment variables"""

    DB_URI_ENV_VAR = "VANTAGE6_ALGO_STORE_DB_URI"
    CONFIG_NAME_ENV_VAR = "VANTAGE6_ALGO_STORE_CONFIG_NAME"



import click
import json
import subprocess
import sys

from pathlib import Path

from rich.console import Console

from vantage6.cli.dev.create import create_demo_network
from vantage6.cli.dev.remove import remove_demo_network
from vantage6.cli.dev.start import start_demo_network
from vantage6.cli.dev.stop import stop_demo_network
from vantage6.cli.utils import prompt_config_name
from vantage6.common.globals import Ports

TEST_FILE_PATH = Path(__file__).parent / "algo_test_scripts" / "algo_test_script.py"


@click.command()
@click.option(
    "--script",
    type=click.Path(),
    default=TEST_FILE_PATH,
    help="Path of the script to test the algorithm. If a script is not provided, the default script is used.",
)
@click.option(
    "--task-arguments",
    type=str,
    default=None,
    help="Arguments to be provided to Task.create function. If --script is provided, this should not be set.",
)
@click.option(
    "--create-dev-network",
    is_flag=True,
    help="Create a new dev network to run the test",
)
@click.option(
    "--start-dev-network",
    is_flag=True,
    help="Start a dev network to run the test",
)
@click.option(
    "-n", "--name", default=None, type=str, help="Name for your development setup"
)
@click.option(
    "--server-url",
    type=str,
    default="http://host.docker.internal",
    help="Server URL to point to. If you are using Docker Desktop, "
    "the default http://host.docker.internal should not be changed.",
)
@click.option(
    "-i", "--image", type=str, default=None, help="Server Docker image to use"
)
@click.option(
    "--keep",
    type=bool,
    default=False,
    help="Keep the dev network after finishing the test",
)
@click.option(
    "--add-dataset",
    type=(str, click.Path()),
    default=[],
    multiple=True,
    help="Add a dataset to the nodes. The first argument is the label of the database, "
    "the second is the path to the dataset file.",
)
@click.pass_context
def cli_test_client_script(
    click_ctx: click.Context,
    script: Path | None,
    task_arguments: str | None,
    name: str,
    server_url: str,
    create_dev_network: bool,
    start_dev_network: bool,
    image: str,
    keep: bool,
    add_dataset: list[tuple[str, Path]] = (),
) -> int:
    """
    Run a script for testing an algorithm on a dev network.
    The path to the script must be provided as an argument.
    """
    if not (script or task_arguments):
        raise click.UsageError("--script or --task-arguments must be set.")
    elif script != TEST_FILE_PATH and task_arguments:
        raise click.UsageError("--script and --task-arguments cannot be set together.")

    # Check if the task_arguments is a valid JSON string
    if task_arguments:
        try:
            json.loads(task_arguments.replace("'", '"'))
        except json.JSONDecodeError:
            raise click.UsageError("task-arguments must be a valid JSON string.")

    name = prompt_config_name(name)

    # create the network
    if create_dev_network:
        click_ctx.invoke(
            create_demo_network,
            name=name,
            num_nodes=3,
            server_url=server_url,
            server_port=Ports.DEV_SERVER.value,
            image=image,
            extra_server_config=None,
            extra_node_config=None,
            add_dataset=add_dataset,
        )

    # start the server and nodes
    if create_dev_network or start_dev_network:
        click_ctx.invoke(
            start_demo_network,
            name=name,
            server_image=image,
            node_image=image,
        )

    # run the test script and get the result
    if not task_arguments:
        subprocess_args = ["python", script]
    else:
        subprocess_args = ["python", script, task_arguments]

    result = subprocess.run(subprocess_args, stdout=sys.stdout, stderr=sys.stderr)

    # check the exit code. If the test passed, it should be 0
    if result.returncode == 0:
        msg = ":heavy_check_mark: [green]Test passed[/green]"
    else:
        msg = ":x: [red]Test failed[/red]"

    console = Console()
    console.print(msg)

    # clean up the test resources. Keep the network if --keep is set, or if the network
    # was created for this test. If the network was started for this test, stop it but
    # do not remove it.
    if not keep:
        if create_dev_network or start_dev_network:
            click_ctx.invoke(stop_demo_network, name=name)
        if create_dev_network:
            click_ctx.invoke(remove_demo_network, name=name)

    return result.returncode



import yaml
import docker
from pathlib import Path

from vantage6.common import info, error
from vantage6.common.docker.network_manager import NetworkManager
from vantage6.common.globals import DEFAULT_PROMETHEUS_EXPORTER_PORT
from vantage6.cli.context.server import ServerContext
from vantage6.cli.globals import (
    DEFAULT_PROMETHEUS_IMAGE,
    PROMETHEUS_CONFIG,
)


class PrometheusServer:
    """
    Manages the Prometheus Docker container
    """

    def __init__(
        self, ctx: ServerContext, network_mgr: NetworkManager, image: str = None
    ):
        """
        Initialize the PrometheusServer instance.

        Parameters
        ----------
        ctx : ServerContext
            The server context containing configuration and paths.
        network_mgr : NetworkManager
            The network manager responsible for managing Docker networks.
        image : str, optional
            The Docker image to use for the Prometheus container. If not provided,
            the default Prometheus image will be used.
        """
        self.ctx = ctx
        self.network_mgr = network_mgr
        self.docker = docker.from_env()
        self.image = image if image else DEFAULT_PROMETHEUS_IMAGE
        self.config_file = Path(__file__).parent / PROMETHEUS_CONFIG
        self.data_dir = self.ctx.prometheus_dir

    def start(self):
        """
        Start a Docker container running Prometheus
        """
        self._prepare_config()

        volumes = {
            str(self.config_file): {
                "bind": "/etc/prometheus/prometheus.yml",
                "mode": "ro",
            },
            str(self.data_dir): {"bind": "/prometheus", "mode": "rw"},
        }
        ports = {"9090/tcp": 9090}

        if self._is_container_running():
            info("Prometheus is already running!")
            return

        self.docker.containers.run(
            name=self.ctx.prometheus_container_name,
            image=self.image,
            volumes=volumes,
            ports=ports,
            detach=True,
            restart_policy={"Name": "unless-stopped"},
            network=self.network_mgr.network_name,
        )
        info("Prometheus container started successfully!")

    def _prepare_config(self):
        """
        Prepare the Prometheus configuration and data directories
        """
        if not self.config_file.exists():
            error(f"Prometheus configuration file {self.config_file} not found!")
            raise FileNotFoundError(f"{self.config_file} not found!")

        if not self.data_dir.exists():
            self.data_dir.mkdir(parents=True, exist_ok=True)

        self._update_prometheus_config()

    def _update_prometheus_config(self):
        """
        Update the Prometheus configuration file with the server address.
        """

        try:
            prometheus_exporter_port = self.ctx.config.get("prometheus", {}).get(
                "exporter_port", DEFAULT_PROMETHEUS_EXPORTER_PORT
            )
            server_hostname = self.ctx.prometheus_container_name
            server_address = f"{server_hostname}:{prometheus_exporter_port}"

            info(
                f"Using Docker container hostname '{server_hostname}' for Prometheus "
                "target. Ensure Prometheus is in the same Docker network to resolve "
                "this address."
            )

            with open(self.config_file, "r") as f:
                config = yaml.safe_load(f)

            job_name = "vantage6_server_metrics"
            job_exists = any(
                job.get("job_name") == job_name
                for job in config.get("scrape_configs", [])
            )

            if not job_exists:
                new_job = {
                    "job_name": job_name,
                    "static_configs": [{"targets": [server_address]}],
                }
                config.setdefault("scrape_configs", []).append(new_job)
            else:
                for job in config["scrape_configs"]:
                    if job.get("job_name") == job_name:
                        job["static_configs"] = [{"targets": [server_address]}]

            with open(self.config_file, "w") as f:
                yaml.dump(config, f)

            info(f"Prometheus configuration updated with target: {server_address}")

        except Exception as e:
            error(f"Failed to update Prometheus configuration: {e}")
            raise

    def _is_container_running(self) -> bool:
        """
        Check if a Prometheus container is already running.

        Returns
        -------
        bool
            True if the Prometheus container is running, False otherwise.
        """
        try:
            container = self.docker.containers.get(self.ctx.prometheus_container_name)
            return container.status == "running"
        except docker.errors.NotFound:
            return False



import click
import docker
from docker.client import DockerClient

from vantage6.common import info, warning, error
from vantage6.common.docker.network_manager import NetworkManager
from vantage6.common.globals import (
    APPNAME,
    DEFAULT_SERVER_IMAGE,
    DEFAULT_UI_IMAGE,
    InstanceType,
)

from vantage6.common.globals import Ports, DEFAULT_PROMETHEUS_EXPORTER_PORT
from vantage6.cli.context.server import ServerContext
from vantage6.cli.rabbitmq.queue_manager import RabbitMQManager
from vantage6.cli.server.common import stop_ui
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.prometheus.monitoring_manager import PrometheusServer
from vantage6.cli.common.start import (
    attach_logs,
    check_for_start,
    get_image,
    mount_database,
    mount_source,
    pull_infra_image,
)


@click.command()
@click.option("--ip", default=None, help="IP address to listen on")
@click.option("-p", "--port", default=None, type=int, help="Port to listen on")
@click.option("-i", "--image", default=None, help="Server Docker image to use")
@click.option(
    "--with-ui",
    "start_ui",
    flag_value=True,
    default=False,
    help="Start the graphical User Interface as well",
)
@click.option(
    "--ui-port", default=None, type=int, help="Port to listen on for the User Interface"
)
@click.option(
    "--with-rabbitmq",
    "start_rabbitmq",
    flag_value=True,
    default=False,
    help="Start RabbitMQ message broker as local "
    "container - use in development only",
)
@click.option("--rabbitmq-image", default=None, help="RabbitMQ docker image to use")
@click.option(
    "--with-prometheus",
    "start_prometheus",
    flag_value=True,
    default=False,
    help="Start Prometheus monitoring as a local container",
)
@click.option("--prometheus-image", default=None, help="Prometheus docker image to use")
@click.option(
    "--keep/--auto-remove",
    default=False,
    help="Keep image after server has stopped. Useful for debugging",
)
@click.option(
    "--mount-src",
    default="",
    help="Override vantage6 source code in container with the source"
    " code in this path",
)
@click.option(
    "--attach/--detach",
    default=False,
    help="Print server logs to the console after start",
)
@click_insert_context(type_="server")
def cli_server_start(
    ctx: ServerContext,
    ip: str,
    port: int,
    image: str,
    start_ui: bool,
    ui_port: int,
    start_rabbitmq: bool,
    rabbitmq_image: str,
    start_prometheus: bool,
    prometheus_image: str,
    keep: bool,
    mount_src: str,
    attach: bool,
) -> None:
    """
    Start the server.
    """
    info("Starting server...")
    docker_client = check_for_start(ctx, InstanceType.SERVER.value)

    image = get_image(image, ctx, "server", DEFAULT_SERVER_IMAGE)

    # check that log directory exists - or create it
    ctx.log_dir.mkdir(parents=True, exist_ok=True)

    info("Pulling server image...")
    pull_infra_image(docker_client, image, InstanceType.SERVER)

    info("Creating mounts")
    config_file = "/mnt/config.yaml"
    mounts = [
        docker.types.Mount(config_file, str(ctx.config_file), type="bind"),
        docker.types.Mount("/mnt/log/", str(ctx.log_dir), type="bind"),
    ]

    src_mount = mount_source(mount_src)
    if src_mount:
        mounts.append(src_mount)

    mount, environment_vars = mount_database(ctx, InstanceType.SERVER)
    if mount:
        mounts.append(mount)

    # Create a docker network for the server and other services like RabbitMQ
    # to reside in
    server_network_mgr = NetworkManager(
        network_name=f"{APPNAME}-{ctx.name}-{ctx.scope}-network"
    )
    server_network_mgr.create_network(is_internal=False)

    if (
        start_rabbitmq
        or ctx.config.get("rabbitmq")
        and ctx.config["rabbitmq"].get("start_with_server", False)
    ):
        # Note that ctx.data_dir has been created at this point, which is
        # required for putting some RabbitMQ configuration files inside
        info("Starting RabbitMQ container")
        _start_rabbitmq(ctx, rabbitmq_image, server_network_mgr)
    elif ctx.config.get("rabbitmq"):
        info(
            "RabbitMQ is provided in the config file as external service. "
            "Assuming this service is up and running."
        )
    else:
        warning(
            "Message queue is not set up! This means that the vantage6 server "
            "cannot be scaled horizontally!"
        )

    if (
        start_prometheus
        or ctx.config.get("prometheus")
        and ctx.config["prometheus"].get("start_with_server", False)
    ):
        info("Starting Prometheus container")
        _start_prometheus(ctx, prometheus_image, server_network_mgr)
    elif ctx.config.get("prometheus", {}).get("enabled"):
        info(
            "Prometheus is enabled but is set not to be started by vantage6. "
            "Assuming this service is up and running."
        )
    else:
        warning(
            "Monitoring is not set up! This means that the vantage6 server "
            "cannot be monitored with Prometheus!"
        )

    # start the UI if requested
    if start_ui or ctx.config.get("ui") and ctx.config["ui"].get("enabled"):
        _start_ui(docker_client, ctx, ui_port)

    # The `ip` and `port` refer here to the ip and port within the container.
    # So we do not really care that is it listening on all interfaces.
    internal_port = 5000
    cmd = (
        f"uwsgi --http :{internal_port} --gevent 1000 --http-websockets "
        "--http-chunked-input --http-keepalive --post-buffering 0 "
        "--master --callable app --disable-logging "
        "--wsgi-file /vantage6/vantage6-server/vantage6/server/wsgi.py "
        f"--pyargv {config_file}"
    )

    info(cmd)

    info("Run Docker container")
    port_ = str(port or ctx.config["port"] or Ports.DEV_SERVER.value)
    prometheus_exporter_port = ctx.config.get("prometheus", {}).get(
        "exporter_port", DEFAULT_PROMETHEUS_EXPORTER_PORT
    )
    container = docker_client.containers.run(
        image,
        command=cmd,
        mounts=mounts,
        detach=True,
        labels={
            f"{APPNAME}-type": InstanceType.SERVER.value,
            "name": ctx.config_file_name,
        },
        environment=environment_vars,
        ports={
            f"{internal_port}/tcp": (ip, port_),  # API port
            f"{prometheus_exporter_port}/tcp": prometheus_exporter_port,
        },
        name=ctx.docker_container_name,
        auto_remove=not keep,
        tty=True,
        network=server_network_mgr.network_name,
    )

    info(f"Success! container id = {container.id}")

    if attach:
        attach_logs(container, InstanceType.SERVER)


def _start_rabbitmq(
    ctx: ServerContext, rabbitmq_image: str, network_mgr: NetworkManager
) -> None:
    """
    Start the RabbitMQ container if it is not already running.

    Parameters
    ----------
    ctx : ServerContext
        Server context object
    rabbitmq_image : str
        RabbitMQ image to use
    network_mgr : NetworkManager
        Network manager object
    """
    rabbit_uri = ctx.config["rabbitmq"].get("uri")
    if not rabbit_uri:
        error(
            "No RabbitMQ URI found in the configuration file! Please add"
            "a 'uri' key to the 'rabbitmq' section of the configuration."
        )
        exit(1)
    # kick off RabbitMQ container
    rabbit_mgr = RabbitMQManager(ctx=ctx, network_mgr=network_mgr, image=rabbitmq_image)
    rabbit_mgr.start()


def _start_prometheus(
    ctx: ServerContext, prometheus_image: str, network_mgr: NetworkManager
) -> None:
    """
    Start the Prometheus container if it is not already running.

    Parameters
    ----------
    ctx : ServerContext
        Server context object
    prometheus_image : str
        Prometheus image to use
    network_mgr : NetworkManager
        Network manager object
    """
    prometheus_server = PrometheusServer(
        ctx=ctx, network_mgr=network_mgr, image=prometheus_image
    )
    prometheus_server.start()


def _start_ui(client: DockerClient, ctx: ServerContext, ui_port: int) -> None:
    """
    Start the UI container.

    Parameters
    ----------
    client : DockerClient
        Docker client
    ctx : ServerContext
        Server context object
    ui_port : int
        Port to expose the UI on
    """
    # if no port is specified, check if config contains a port
    ui_config = ctx.config.get("ui")
    if ui_config and not ui_port:
        ui_port = ui_config.get("port")

    # check if the port is valid
    # TODO make function to check if port is valid, and use in more places
    if not isinstance(ui_port, int) or not 0 < ui_port < 65536:
        warning(
            f"UI port '{ui_port}' is not valid! Using default port "
            f"{Ports.DEV_UI.value}"
        )
        ui_port = str(Ports.DEV_UI.value)

    # find image to use
    image = get_image(None, ctx, "ui", DEFAULT_UI_IMAGE)

    pull_infra_image(client, image, InstanceType.UI)

    # set environment variables
    env_vars = {
        "SERVER_URL": f"http://localhost:{ctx.config.get('port')}",
        "API_PATH": ctx.config.get("api_path"),
    }

    # stop the UI container if it is already running
    stop_ui(client, ctx)

    info(f"Starting User Interface at port {ui_port}")
    ui_container_name = f"{APPNAME}-{ctx.name}-{ctx.scope}-ui"
    client.containers.run(
        image,
        detach=True,
        labels={f"{APPNAME}-type": "ui", "name": ctx.config_file_name},
        ports={"80/tcp": (ctx.config.get("ip"), ui_port)},
        name=ui_container_name,
        environment=env_vars,
        tty=True,
    )



from __future__ import annotations

import os.path

from pathlib import Path

from vantage6.common.context import AppContext
from vantage6.common.globals import APPNAME, InstanceType
from vantage6.cli.configuration_manager import NodeConfigurationManager
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli._version import __version__


class NodeContext(AppContext):
    """
    Node context object for the host system.

    See DockerNodeContext for the node instance mounts when running as a
    dockerized service.

    Parameters
    ----------
    instance_name : str
        Name of the configuration instance, corresponds to the filename
        of the configuration file.
    system_folders : bool, optional
        _description_, by default N_FOL
    config_file : str, optional
        _description_, by default None
    """

    # The server configuration manager is aware of the structure of the server
    # configuration file and makes sure only valid configuration can be loaded.
    INST_CONFIG_MANAGER = NodeConfigurationManager

    def __init__(
        self,
        instance_name: str,
        system_folders: bool = N_FOL,
        config_file: str = None,
        print_log_header: bool = True,
    ):
        super().__init__(
            InstanceType.NODE,
            instance_name,
            system_folders,
            config_file,
            print_log_header,
        )
        if print_log_header:
            self.log.info("vantage6 version '%s'", __version__)

    @classmethod
    def from_external_config_file(
        cls, path: str, system_folders: bool = N_FOL
    ) -> NodeContext:
        """
        Create a node context from an external configuration file. External
        means that the configuration file is not located in the default folders
        but its location is specified by the user.

        Parameters
        ----------
        path : str
            Path of the configuration file
        system_folders : bool, optional
            System wide or user configuration, by default N_FOL

        Returns
        -------
        NodeContext
            Node context object
        """
        return super().from_external_config_file(
            Path(path).resolve(), InstanceType.NODE, system_folders
        )

    @classmethod
    def config_exists(cls, instance_name: str, system_folders: bool = N_FOL) -> bool:
        """
        Check if a configuration file exists.

        Parameters
        ----------
        instance_name : str
            Name of the configuration instance, corresponds to the filename
            of the configuration file.
        system_folders : bool, optional
            System wide or user configuration, by default N_FOL

        Returns
        -------
        bool
            Whether the configuration file exists or not
        """
        return super().config_exists(
            InstanceType.NODE, instance_name, system_folders=system_folders
        )

    @classmethod
    def available_configurations(
        cls, system_folders: bool = N_FOL
    ) -> tuple[list, list]:
        """
        Find all available server configurations in the default folders.

        Parameters
        ----------
        system_folders : bool, optional
            System wide or user configuration, by default N_FOL

        Returns
        -------
        tuple[list, list]
            The first list contains validated configuration files, the second
            list contains invalid configuration files.
        """
        return super().available_configurations(InstanceType.NODE, system_folders)

    @staticmethod
    def type_data_folder(system_folders: bool = N_FOL) -> Path:
        """
        Obtain OS specific data folder where to store node specific data.

        Parameters
        ----------
        system_folders : bool, optional
            System wide or user configuration

        Returns
        -------
        Path
            Path to the data folder
        """
        return AppContext.type_data_folder(InstanceType.NODE.value, system_folders)

    @property
    def databases(self) -> dict:
        """
        Dictionary of local databases that are available for this node.

        Returns
        -------
        dict
            dictionary with database names as keys and their corresponding
            paths as values.
        """
        return self.config["databases"]

    @property
    def docker_container_name(self) -> str:
        """
        Docker container name of the node.

        Returns
        -------
        str
            Node's Docker container name
        """
        return f"{APPNAME}-{self.name}-{self.scope}"

    @property
    def docker_network_name(self) -> str:
        """
        Private Docker network name which is unique for this node.

        Returns
        -------
        str
            Docker network name
        """
        return f"{APPNAME}-{self.name}-{self.scope}-net"

    @property
    def docker_volume_name(self) -> str:
        """
        Docker volume in which task data is stored. In case a file based
        database is used, this volume contains the database file as well.

        Returns
        -------
        str
            Docker volume name
        """
        return os.environ.get("DATA_VOLUME_NAME", f"{self.docker_container_name}-vol")

    @property
    def docker_vpn_volume_name(self) -> str:
        """
        Docker volume in which the VPN configuration is stored.

        Returns
        -------
        str
            Docker volume name
        """
        return os.environ.get(
            "VPN_VOLUME_NAME", f"{self.docker_container_name}-vpn-vol"
        )

    @property
    def docker_ssh_volume_name(self) -> str:
        """
        Docker volume in which the SSH configuration is stored.

        Returns
        -------
        str
            Docker volume name
        """
        return os.environ.get(
            "SSH_TUNNEL_VOLUME_NAME", f"{self.docker_container_name}-ssh-vol"
        )

    @property
    def docker_squid_volume_name(self) -> str:
        """
        Docker volume in which the SSH configuration is stored.

        Returns
        -------
        str
            Docker volume name
        """
        return os.environ.get(
            "SSH_SQUID_VOLUME_NAME", f"{self.docker_container_name}-squid-vol"
        )

    @property
    def proxy_log_file(self):
        return self.log_file_name(type_="proxy_server")

    def docker_temporary_volume_name(self, job_id: int) -> str:
        """
        Docker volume in which temporary data is stored. Temporary data is
        linked to a specific run. Multiple algorithm containers can have the
        same run id, and therefore the share same temporary volume.

        Parameters
        ----------
        job_id : int
            run id provided by the server

        Returns
        -------
        str
            Docker volume name
        """
        return f"{APPNAME}-{self.name}-{self.scope}-{job_id}-tmpvol"

    def get_database_uri(self, label: str = "default") -> str:
        """
        Obtain the database URI for a specific database.

        Parameters
        ----------
        label : str, optional
            Database label, by default "default"

        Returns
        -------
        str
            URI to the database
        """
        return self.config["databases"][label]



from vantage6.common.globals import InstanceType
from vantage6.cli.configuration_wizard import select_configuration_questionaire
from vantage6.cli.context import get_context
from vantage6.cli.context.server import ServerContext


def get_dev_server_context(config: str | None, name: str | None) -> ServerContext:
    """
    Get the server context for the development server.

    Parameters
    ----------
    config : str | None
        Path to the configuration file. If None, the name will be used.
    name : str | None
        Name of the configuration. If None, a questionaire will be shown.
    """
    if config:
        return ServerContext.from_external_config_file(config)
    if not name:
        name = select_configuration_questionaire(
            InstanceType.SERVER, system_folders=False
        )
    return get_context(InstanceType.SERVER, name, system_folders=False)



import click
import questionary as q
import docker

from vantage6.common import error
from vantage6.common.globals import APPNAME
from vantage6.common.docker.addons import check_docker_running
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli import __version__
from vantage6.cli.node.common import find_running_node_names


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for configuration in system folders rather than " "user folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for configuration in user folders rather than "
    "system folders. This is the default",
)
def cli_node_version(name: str, system_folders: bool) -> None:
    """
    Returns current version of a vantage6 node.
    """
    check_docker_running()
    client = docker.from_env()

    running_node_names = find_running_node_names(client)

    if not name:
        if not running_node_names:
            error(
                "No nodes are running! You can only check the version for "
                "nodes that are running"
            )
            exit(1)
        try:
            name = q.select(
                "Select the node you wish to inspect:", choices=running_node_names
            ).unsafe_ask()
        except KeyboardInterrupt:
            error("Aborted by user!")
            return
    else:
        post_fix = "system" if system_folders else "user"
        name = f"{APPNAME}-{name}-{post_fix}"

    if name in running_node_names:
        container = client.containers.get(name)
        version = container.exec_run(cmd="vnode-local version", stdout=True)
        click.echo({"node": version.output.decode("utf-8"), "cli": __version__})
    else:
        error(f"Node {name} is not running! Cannot provide version...")



import subprocess
import click
import questionary as q
import docker

from vantage6.common import warning, error
from vantage6.common.docker.addons import check_docker_running
from vantage6.cli.common.utils import get_name_from_container_name
from vantage6.cli.node.stop import cli_node_stop
from vantage6.cli.node.common import find_running_node_names
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for configuration in system folders instead of " "user folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for configuration in the user folders instead of "
    "system folders. This is the default.",
)
@click.option("-i", "--image", default=None, help="Node Docker image to use")
@click.option(
    "--keep/--auto-remove",
    default=False,
    help="Keep node container after finishing. Useful for debugging",
)
@click.option(
    "--force-db-mount",
    is_flag=True,
    help="Always mount node databases; skip the check if they are existing files.",
)
@click.option(
    "--attach/--detach",
    default=False,
    help="Show node logs on the current console after starting the " "node",
)
@click.option(
    "--mount-src",
    default="",
    help="Override vantage6 source code in container with the source"
    " code in this path",
)
@click.option("--all", "all_nodes", flag_value=True, help="Stop all running nodes")
@click.option(
    "--force",
    "force",
    flag_value=True,
    help="Kill nodes instantly; don't wait for them to shut down",
)
@click.pass_context
def cli_node_restart(
    click_ctx: click.Context,
    name: str,
    system_folders: bool,
    image: str,
    keep: bool,
    mount_src: str,
    attach: bool,
    force_db_mount: bool,
    all_nodes: bool,
    force: bool,
) -> None:
    """Restart the node"""
    check_docker_running()
    client = docker.from_env()

    running_node_names = find_running_node_names(client)
    if not running_node_names:
        warning("No nodes are currently running. No action taken.")
        return

    if attach and all_nodes:
        error(
            "Cannot attach logs of all nodes at once. Please remove either the "
            "'--all' or '--attach' option."
        )
        return

    if all_nodes:
        names = [
            get_name_from_container_name(container_name)
            for container_name in running_node_names
        ]
    else:
        if not name:
            try:
                container_name = q.select(
                    "Select the node you wish to restart:", choices=running_node_names
                ).unsafe_ask()
            except KeyboardInterrupt:
                error("Aborted by user!")
                return
            names = [get_name_from_container_name(container_name)]
        else:
            names = [name]

    for node_name in names:
        click_ctx.invoke(
            cli_node_stop,
            name=node_name,
            system_folders=system_folders,
            all_nodes=False,
            force=force,
        )

        cmd = ["v6", "node", "start", "--name", node_name]
        if system_folders:
            cmd.append("--system")
        if image:
            cmd.extend(["--image", image])
        if keep:
            cmd.append("--keep")
        if mount_src:
            cmd.extend(["--mount-src", mount_src])
        if attach:
            cmd.append("--attach")
        if force_db_mount:
            cmd.append("--force-db-mount")
        subprocess.run(cmd, check=True)



import codecs
import os
from codecs import open
from os import path
from pathlib import Path

from setuptools import find_namespace_packages, setup

# get current directory
here = Path(path.abspath(path.dirname(__file__)))
parent_dir = here.parent.absolute()

# get the long description from the README file
with codecs.open(path.join(parent_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Read the API version from disk. This file should be located in the package
# folder, since it's also used to set the pkg.__version__ variable.
version_path = os.path.join(here, "vantage6", "cli", "_version.py")
version_ns = {"__file__": version_path}
with open(version_path) as f:
    exec(f.read(), {}, version_ns)


# setup the package
setup(
    name="vantage6",
    version=version_ns["__version__"],
    description="vantage6 command line interface",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vantage6/vantage6",
    packages=find_namespace_packages(),
    python_requires=">=3.10",
    install_requires=[
        "click==8.1.3",
        "colorama==0.4.6",
        "copier==9.9.1",
        "docker==7.1.0",
        "ipython==8.10.0",
        "jinja2==3.1.6",
        "pandas>=1.5.3",
        "questionary==2.1.1",
        "rich==13.5.2",
        "schema==0.7.5",
        "SQLAlchemy==1.4.46",
        f"vantage6-common == {version_ns['__version__']}",
        f"vantage6-client == {version_ns['__version__']}",
    ],
    extras_require={
        "dev": [
            "coverage==6.4.4",
            "black",
            "pre-commit",
        ]
    },
    package_data={
        "vantage6.cli": [
            "__build__",
            "rabbitmq/rabbitmq.config",
            "template/node_config.j2",
            "template/server_config.j2",
            "template/server_import_config.j2",
            "template/algo_store_config.j2",
            "dev/data/olympic_athletes_2016.csv",
            "dev/data/km_dataset.csv",
            "prometheus/prometheus.yml",
        ],
    },
    entry_points={
        "console_scripts": [
            "v6=vantage6.cli.cli:cli_complete",
            # TODO v5+ remove vnode/vserver/vdev aliases (?)
            "vnode=vantage6.cli.cli:cli_node",
            "vserver=vantage6.cli.cli:cli_server",
            "vdev=vantage6.cli.cli:cli_dev",
        ]
    },
)



import os
import json

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "__build__")) as fp:
    __build__ = json.load(fp)

# Module version
version_info = (4, 13, 4, "candidate", __build__, 0)

# Module version stage suffix map
_specifier_ = {"alpha": "a", "beta": "b", "candidate": "rc", "final": ""}
version = f"{version_info[0]}.{version_info[1]}.{version_info[2]}"
pre_release = (
    ""
    if version_info[3] == "final"
    else _specifier_[version_info[3]] + str(version_info[4])
)
post_release = "" if not version_info[5] else f".post{version_info[5]}"

# Module version accessible using thomas.__version__
__version__ = f"{version}{pre_release}{post_release}"
